----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date: 18.04.2021 21:19:41
-- Design Name: 
-- Module Name: Etage1_LectureInstruction - Behavioral
-- Project Name: 
-- Target Devices: 
-- Tool Versions: 
-- Description: 
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity Etage1_LectureInstruction is
    Generic (Instruction_size_in_memory : Natural;
             Addr_size_mem_instruction : Natural;
             Mem_instruction_size : Natural;
             Nb_bits : Natural;
             Instruction_bus_size : Natural;
             Nb_registres : Natural;
             Mem_adresse_retour_size : Natural;
             Adresse_size_mem_adresse_retour : Natural;
             Instructions_critiques_lecture_A : STD_LOGIC_VECTOR;
             Instructions_critiques_lecture_B : STD_LOGIC_VECTOR;
             Instructions_critiques_lecture_C : STD_LOGIC_VECTOR;
             Instructions_critiques_ecriture : STD_LOGIC_VECTOR;
             Code_Instruction_JMP : STD_LOGIC_VECTOR;
             Code_Instruction_JMZ : STD_LOGIC_VECTOR;
             Code_Instruction_CALL : STD_LOGIC_VECTOR;
             Code_Instruction_RET : STD_LOGIC_VECTOR;
             Code_Instruction_STOP : STD_LOGIC_VECTOR);
    Port ( CLK : in STD_LOGIC;
           RST : in STD_LOGIC;
           Z : in STD_LOGIC;
           A : out STD_LOGIC_VECTOR (Nb_bits - 1 downto 0);
           B : out STD_LOGIC_VECTOR (Nb_bits - 1 downto 0);
           C : out STD_LOGIC_VECTOR (Nb_bits - 1 downto 0);
           Instruction : out STD_LOGIC_VECTOR (Instruction_bus_size - 1 downto 0));
end Etage1_LectureInstruction;

architecture Behavioral of Etage1_LectureInstruction is
    component MemoireInstructions is
    Generic (Nb_bits : Natural;
             Addr_size : Natural;
             Mem_size : Natural);
    Port ( Addr : in STD_LOGIC_VECTOR (Addr_size-1 downto 0);
           D_OUT : out STD_LOGIC_VECTOR (Nb_bits-1 downto 0) := (others => '0'));
    end component;
    
    component MemoireAdressesRetour is
        Generic (Nb_bits : Natural;
                 Addr_size : Natural;
                 Mem_size : Natural);
        Port ( R : in STD_LOGIC;
               W : in STD_LOGIC;
               D_IN : in STD_LOGIC_VECTOR (Nb_bits-1 downto 0);
               RST : in STD_LOGIC;
               CLK : in STD_LOGIC;
               D_OUT : out STD_LOGIC_VECTOR (Nb_bits-1 downto 0) := (others => '0');
               E : out STD_LOGIC;
               F : out STD_LOGIC);
    end component;
    
    signal Pointeur_instruction : STD_LOGIC_VECTOR (Addr_size_mem_instruction - 1 downto 0) := (others => '0');
    signal Pointeur_instruction_next : STD_LOGIC_VECTOR (Addr_size_mem_instruction - 1 downto 0) := (0 => '1', others => '0');
    signal Instruction_courante : STD_LOGIC_VECTOR (Instruction_size_in_memory - 1 downto 0) := (others => '0');
    
    subtype Registre is integer range -1 to Nb_registres - 1;
    type Tab_registres is array (1 to 3) of Registre;
    signal Tableau : Tab_registres := (others => - 1);
    
    signal Adresse_Retour : STD_LOGIC_VECTOR (Addr_size_mem_instruction - 1 downto 0) := (others => '0');
    signal E : STD_LOGIC;
    signal F : STD_LOGIC;
    signal R_Aux : STD_LOGIC := '0';
    signal W_Aux : STD_LOGIC := '0';

    constant Instruction_nulle : STD_LOGIC_VECTOR (Instruction_bus_size - 1 downto 0) := (others => '0');
    constant Argument_nul : STD_LOGIC_VECTOR (Nb_bits - 1 downto 0) := (others => '0');
    
    signal bulles : boolean := false;
    
    signal compteur : integer := 0;
    signal locked : boolean := false;
    
begin
    instance : MemoireInstructions
    generic map (Nb_bits => Instruction_size_in_memory,
                 Addr_size => Addr_size_mem_instruction,
                 Mem_size => Mem_instruction_size)
    port map (Addr => Pointeur_Instruction,
              D_OUT => Instruction_courante);
              
    instance_MemoireAdressesRetour : MemoireAdressesRetour
    generic map (Nb_bits => Addr_size_mem_instruction,
                 Addr_size => Adresse_size_mem_adresse_retour,
                 Mem_size => Mem_adresse_retour_size
    )
    port map ( R => R_Aux,
               W => W_Aux,
               D_IN => Pointeur_instruction_next,
               RST => RST,
               CLK => CLK,
               D_OUT => Adresse_Retour,
               E => E,
               F => F
    );

              
    process 
    begin
        wait until CLK'event and CLK = '1';
        if (RST = '0') then
            Tableau <= (others => -1);
            Pointeur_Instruction <= (others => '0');
            compteur <= 0;
            locked <= false;
            C <= Argument_nul;
            B <= Argument_nul;
            A <= Argument_nul;
            Instruction <= Instruction_nulle;
        else
            Tableau(3) <= Tableau(2);
            Tableau(2) <= Tableau(1);
            Tableau(1) <= -1;
            if (not bulles) then
                if ((Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits) = Code_Instruction_CALL) or (Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits) = Code_Instruction_JMP)) then
                    C <= Instruction_courante (1 * Nb_bits - 1 downto 0 * Nb_bits);
                    B <= Instruction_courante (2 * Nb_bits - 1 downto 1 * Nb_bits);
                    A <= Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits);
                    Instruction <= Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits);
                    Pointeur_Instruction <= Instruction_courante (2 * Nb_bits + Addr_size_mem_instruction - 1 downto 2 * Nb_bits);
                elsif (Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits) = Code_Instruction_RET) then
                    C <= Instruction_courante (1 * Nb_bits - 1 downto 0 * Nb_bits);
                    B <= Instruction_courante (2 * Nb_bits - 1 downto 1 * Nb_bits);
                    A <= Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits);
                    Instruction <= Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits);
                    Pointeur_Instruction <= Adresse_Retour;
                elsif (Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits) = Code_Instruction_JMZ) then
                    compteur <= compteur + 1;
                    C <= Instruction_courante (1 * Nb_bits - 1 downto 0 * Nb_bits);
                    B <= Instruction_courante (2 * Nb_bits - 1 downto 1 * Nb_bits);
                    A <= Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits);
                    Instruction <= Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits);
                    if (compteur = 2) then
                        if (Z = '1') then
                            Pointeur_Instruction <= Instruction_courante (2 * Nb_bits + Addr_size_mem_instruction - 1 downto 2 * Nb_bits);
                        else 
                            Pointeur_Instruction <= Pointeur_Instruction + 1;
                        end if;
                        compteur <= 0;
                    end if;
                elsif (Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits) = Code_Instruction_STOP) then
                    if (not locked) then
                        if (Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits) = Argument_nul) then
                            locked <= true;
                        end if;
                        compteur <= compteur + 1;
                        if (compteur + 1 = to_integer(unsigned(Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits)))) then
                            Pointeur_Instruction <= Pointeur_Instruction + 1;
                            compteur <= 0;
                        end if;
                    end if;
                    C <= Instruction_courante (1 * Nb_bits - 1 downto 0 * Nb_bits);
                    B <= Instruction_courante (2 * Nb_bits - 1 downto 1 * Nb_bits);
                    A <= Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits);
                    Instruction <= Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits);
                else
                    C <= Instruction_courante (1 * Nb_bits - 1 downto 0 * Nb_bits);
                    B <= Instruction_courante (2 * Nb_bits - 1 downto 1 * Nb_bits);
                    A <= Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits);
                    Instruction <= Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits);
                    if (Instructions_critiques_ecriture(to_integer(unsigned(Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits)))) = '1') then
                        Tableau(1) <= to_integer(unsigned(Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits)));
                    end if;
                    Pointeur_Instruction <= Pointeur_Instruction + 1;
                end if;
            else 
                C <= Argument_nul;
                B <= Argument_nul;
                A <= Argument_nul;
                Instruction <= Instruction_nulle;
            end if;
        end if;
    end process;
    
    
    -- Condition degueu -> Instruction critique en lecture simple qui lit dans B et B dans tableau ou instruction critique en lecture double qui lit dans C et C dans tableau 
    bulles <= 
    (
        (
            Instructions_critiques_lecture_A(to_integer(unsigned(Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits)))) = '1'
        )
        and 
        (
            (to_integer(unsigned(Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits))) = Tableau(1))
            or 
            (to_integer(unsigned(Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits))) = Tableau(2))
            or 
            (to_integer(unsigned(Instruction_courante (3 * Nb_bits - 1 downto 2 * Nb_bits))) = Tableau(3))
        )
    ) 
    or 
    (
        (
            Instructions_critiques_lecture_B(to_integer(unsigned(Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits)))) = '1'
        )
        and 
        (
            (to_integer(unsigned(Instruction_courante (2 * Nb_bits - 1 downto 1 * Nb_bits))) = Tableau(1))
            or 
            (to_integer(unsigned(Instruction_courante (2 * Nb_bits - 1 downto 1 * Nb_bits))) = Tableau(2))
            or 
            (to_integer(unsigned(Instruction_courante (2 * Nb_bits - 1 downto 1 * Nb_bits))) = Tableau(3))
        )
    ) 
    or
    (
        (
            Instructions_critiques_lecture_C(to_integer(unsigned(Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits)))) = '1'
        )
        and 
        (
            (to_integer(unsigned(Instruction_courante (1 * Nb_bits - 1 downto 0 * Nb_bits))) = Tableau(1))
            or
            (to_integer(unsigned(Instruction_courante (1 * Nb_bits - 1 downto 0 * Nb_bits))) = Tableau(2))
            or 
            (to_integer(unsigned(Instruction_courante (1 * Nb_bits - 1 downto 0 * Nb_bits))) = Tableau(3))
        )
    );
    
    
    R_Aux <= '1' when Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits) = Code_Instruction_RET else
             '0';
    W_Aux <= '1' when Instruction_courante (Instruction_bus_size + 3 * Nb_bits - 1 downto 3 * Nb_bits) = Code_instruction_CALL else
             '0';
    Pointeur_instruction_next <= Pointeur_instruction + 1;
end Behavioral;
